# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

def decodeUtf8String(inputStr):
    '''Convert an UTF8 sequence into a string

    Required for compatibility with Python 2 where str==bytes
    inputStr -- Either a str or bytes instance with UTF8 encoding
    '''
    return inputStr if isinstance(inputStr, str) or not isinstance(inputStr, bytes) else inputStr.decode('utf8')



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

import errno
import os
import json
from freegenes.logger import bot
import sys


################################################################################
## FOLDER OPERATIONS ###########################################################
################################################################################


def mkdir_p(path):
    '''mkdir_p attempts to get the same functionality as mkdir -p
    :param path: the path to create.
    '''
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            bot.error("Error creating path %s, exiting." % path)
            sys.exit(1)


################################################################################
## FILE OPERATIONS #############################################################
################################################################################


def write_file(filename, content, mode="w"):
    '''write_file will open a file, "filename" and write content, "content"
    and properly close the file
    '''
    with open(filename, mode) as filey:
        filey.writelines(content)
    return filename


def write_json(json_obj, filename, mode="w", print_pretty=True):
    '''write_json will (optionally,pretty print) a json object to file
    :param json_obj: the dict to print to json
    :param filename: the output file to write to
    :param pretty_print: if True, will use nicer formatting
    '''
    with open(filename, mode) as filey:
        if print_pretty:
            filey.writelines(
                json.dumps(
                    json_obj,
                    indent=4,
                    separators=(
                        ',',
                        ': ')))
        else:
            filey.writelines(json.dumps(json_obj))
    return filename


def read_file(filename, mode="r", readlines=True):
    '''write_file will open a file, "filename" and write content, "content"
    and properly close the file
    '''
    with open(filename, mode) as filey:
        if readlines:
            content = filey.readlines()
        else:
            content = filey.read()
    return content


def read_json(filename, mode='r'):
    '''read_json reads in a json file and returns
    the data structure as dict.
    '''
    with open(filename, mode) as filey:
        data = json.load(filey)
    return data



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

import os
import re
import semver
from freegenes.logger import bot, decodeUtf8String
import subprocess
import sys
import shlex

################################################################################
# Local commands and requests
################################################################################

def _process_sudo_cmd(cmd, sudo, sudo_options):
    if sudo and sudo_options is not None:
        if isinstance(sudo_options, str):
            sudo_options = shlex.split(sudo_options)
        cmd = ['sudo'] + sudo_options + cmd
    elif sudo:
        cmd = ['sudo'] + cmd
    return cmd


def check_install(software='singularity', quiet=True):
    '''check_install will attempt to run the singularity command, and
       return True if installed. The command line utils will not run 
       without this check.
    '''

    cmd = [software, '--version']
    found = False

    try:
        version = run_command(cmd, quiet=True)
    except: # FileNotFoundError
        return found

    if version is not None:
        if version['return_code'] == 0:
            found = True

        if not quiet:
            version = version['message']
            bot.info("Found %s version %s" % (software.upper(), version))

    return found


def which(software='singularity'):
    '''which returns the full path to where software is installed.
    '''
    cmd = ['which', software]
    result = run_command(cmd, quiet=True)['message'][0]
    return result.strip('\n')
    

def get_singularity_version():
    '''get the full singularity client version as reported by 
       singularity --version [...]. For Singularity 3.x, this means:
       "singularity version 3.0.1-1"
    '''
    version = os.environ.get('SPYTHON_SINGULARITY_VERSION', "")
    if version == "":
        try:
            version = run_command(["singularity", '--version'], quiet=True)
        except: # FileNotFoundError
            return version

        if version['return_code'] == 0:
            if version['message']:
                version = version['message'][0].strip('\n')

    return version

def get_singularity_version_info():
    '''get the full singularity client version as a semantic version"
    '''
    version_string = get_singularity_version()
    prefix = 'singularity version '
    if version_string.startswith(prefix):
        version_string = version_string[len(prefix):]
    elif '/' in version_string: # Handle old stuff like "x.y.z-pull/123-0a5d"
        version_string = version_string.replace('/', '+', 1)
    return semver.parse_version_info(version_string)

def get_installdir():
    '''get_installdir returns the installation directory of the application
    '''
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def stream_command(cmd, no_newline_regexp="Progess", sudo=False, sudo_options=None):
    '''stream a command (yield) back to the user, as each line is available.

       # Example usage:
       results = []
       for line in stream_command(cmd):
           print(line, end="")
           results.append(line)

       Parameters
       ==========
       cmd: the command to send, should be a list for subprocess
       no_newline_regexp: the regular expression to determine skipping a
                          newline. Defaults to finding Progress
       sudo_options: string or list of strings that will be passed as options to sudo

    '''
    cmd = _process_sudo_cmd(cmd, sudo, sudo_options)

    process = subprocess.Popen(cmd,
                               stdout=subprocess.PIPE,
                               universal_newlines=True)
    for line in iter(process.stdout.readline, ""):
        if not re.search(no_newline_regexp, line):
            yield line
    process.stdout.close()
    return_code = process.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def run_command(cmd, 
                sudo=False,
                capture=True,
                no_newline_regexp="Progess",
                quiet=False,
                sudo_options=None):

    '''run_command uses subprocess to send a command to the terminal. If
       capture is True, we use the parent stdout, so the progress bar (and
       other commands of interest) are piped to the user. This means we 
       don't return the output to parse.

       Parameters
       ==========
       cmd: the command to send, should be a list for subprocess
       sudo: if needed, add to start of command
       no_newline_regexp: the regular expression to determine skipping a
                          newline. Defaults to finding Progress
       capture: if True, don't set stdout and have it go to console. This
                option can print a progress bar, but won't return the lines
                as output.
       sudo_options: string or list of strings that will be passed as options to sudo
    '''
    cmd = _process_sudo_cmd(cmd, sudo, sudo_options)

    stdout = None
    if capture:
        stdout = subprocess.PIPE

    # Use the parent stdout and stderr
    process = subprocess.Popen(cmd,
                               stderr=subprocess.PIPE,
                               stdout=stdout)
    lines = []
    found_match = False

    for line in process.communicate():
        if line:
            line = decodeUtf8String(line)
            lines.append(line)
            if re.search(no_newline_regexp, line) and found_match:
                if not quiet:
                    sys.stdout.write(line)
                found_match = True
            else:
                if not quiet:
                    sys.stdout.write(line)
                    print(line.rstrip())
                found_match = False

    output = {'message': lines,
              'return_code': process.returncode}

    return output


################################################################################
# Parsing and Formatting
################################################################################


def format_container_name(name, special_characters=None):
    '''format_container_name will take a name supplied by the user,
    remove all special characters (except for those defined by "special-characters"
    and return the new image name.
    '''
    if special_characters is None:
        special_characters = []
    return ''.join(e.lower()
                   for e in name if e.isalnum() or e in special_characters)


def split_uri(container):
    '''Split the uri of a container into the protocol and image part

    An empty protocol is returned if none found.
    A trailing slash is removed from the image part.
    '''
    parts = container.split('://', 1)
    if len(parts) == 2:
        protocol, image = parts
    else:
        protocol = ''
        image = parts[0]
    return protocol, image.rstrip('/')


def remove_uri(container):
    '''remove_uri will remove docker:// or shub:// from the uri
    '''
    return split_uri(container)[1]




# Copyright (C) 2019 Vanessa Sochat.

# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

__version__ = "0.0.11"
AUTHOR = 'Vanessa Sochat'
AUTHOR_EMAIL = 'vsochat@stanford.edu'
NAME = 'freegenes'
PACKAGE_URL = "http://www.github.com/vsoch/freegenes-python"
KEYWORDS = 'FreeGenes node API client'
DESCRIPTION = "Command line python tool for working with FreeGenes."
LICENSE = "LICENSE"

INSTALL_REQUIRES = (
    ('semver', {'min_version': '2.8.0'}),
    ('requests', {'min_version': '2.21.0'}),
)

TESTS_REQUIRES = (
    ('pytest', {'min_version': '4.6.2'}),
)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

import argparse
import sys
import os


def get_parser():

    parser = argparse.ArgumentParser(description="FreeGenes Python",
                                formatter_class=argparse.RawTextHelpFormatter,
                                add_help=False)

    # Global Options
    parser.add_argument('--debug', '-d', dest="debug", 
                        help="use verbose logging to debug.", 
                        default=False, action='store_true')

    parser.add_argument('--quiet', '-q', dest="quiet", 
                        help="suppress all normal output", 
                        default=False, action='store_true')

    parser.add_argument('--version', dest="version", 
                        help="show version and exit", 
                        default=False, action='store_true')

    subparsers = parser.add_subparsers(help='description',
                                       title='actions',
                                       dest="command", metavar='general usage')

    subparsers.add_parser("shell", help="Interact with freegenes python")
    subparsers.add_parser("twist", help="Interact with twist API")

    return parser


def set_verbosity(args):
    '''determine the message level in the environment to set based on args.
    '''
    level = "INFO"

    if args.debug:
        level = "DEBUG"
    elif args.quiet:
        level = "QUIET"

    os.environ['MESSAGELEVEL'] = level
    os.putenv('MESSAGELEVEL', level)
    
    # Import logger to set
    from freegenes.logger import bot
    bot.debug('Logging level %s' % level)
    import freegenes

    bot.debug("FreeGenes Python Version: %s" % freegenes.__version__)


def version():
    '''version prints the version, both for the user and help output
    '''
    import freegenes
    return freegenes.__version__
    

def main():

    parser = get_parser()

    def print_help(return_code=0):
        '''print help, including the software version and active client 
           and exit with return code.
        '''
        v = version()
        print("\nFreeGenes Python [v%s]\n" %(v))
        parser.print_help()
        sys.exit(return_code)
    
    if len(sys.argv) == 1:
        print_help()
    try:
        # We capture all primary arguments, and take secondary to pass on
        args, options = parser.parse_known_args()
    except:
        sys.exit(0)

    # The main function
    func = None

    # If the user wants the version
    if args.version:
        print(version())
        sys.exit(0)

    # if environment logging variable not set, make silent
    set_verbosity(args)

    # Does the user want help for a subcommand?
    if args.command == 'shell': 
        from .shell import main as func
    elif args.command == 'twist': 
        from .twist import main as func
    else:
        print_help()

    # Pass on to the correct parser
    if args.command is not None:
        func(args=args, options=options, parser=parser)


if __name__ == '__main__':
    main()



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from freegenes.version import __version__
from freegenes.logger import bot
import requests
import os

class Client(object):

    def __init__(self, email=None, token=None, eutoken=None, 
                       base="https://twist-api.twistbioscience-staging.com/", version="v1"):
        '''Generate a client for interacting with Twist.  I was unable to generate
           tokens using the API (it doesn't work), and the head of Twist (Gil Raytan) 
           had to manually send them.

           Endpoints:
             production: https://twist-api.twistdna.com
             staging: https://twist-api.twistbioscience-staging.com/

           Parameters
           ==========
           token: the general api token
           eutoken: the end user token
        '''
        self.version = version
        self._set_base(base)
        self._set_tokens(token, eutoken)
        self._set_headers()
        self._test_token()
        self._set_email(email)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "[client][twist]"

    def _set_tokens(self, token, eutoken):
        '''ensure that token provided, or look for FREEGENES_TWIST_TOKEN 
           or FREEGENES_TWIST_LOGIN and FREEGENES_TWIST_PASSWORD defined in 
           in the environment.
        '''
        self.token = os.environ.get('FREEGENES_TWIST_TOKEN', token)
        self.eutoken = os.environ.get('FREEGENES_TWIST_EUTOKEN', eutoken)

        if not self.token or not self.eutoken:

            username = os.environ.get('FREEGENES_TWIST_LOGIN')
            password = os.environ.get('FREEGENES_TWIST_PASSWORD')

            # username and password are required to get a token
            if not username or not password:
                bot.exit("You must export FREEGENES_TWIST_TOKEN or FREEGENES_TWIST_LOGIN and FREEGENES_TWIST_PASSWORD")

            headers = {"username": username, "password": password}
            response = requests.post(self.base + '/api-token-auth/', headers=headers)
            if response.status_code != 201:
                bot.exit("Error with authentication, %s:%s" %(response.reason, response.status_code))
            self.token = response.json()['token']

    def _set_base(self, base):
        '''look for FREEGENES_TWIST_BASE defined in environ
        '''
        self.base = os.environ.get('FREEGENES_TWIST_BASE', base)
        if self.base:
            self.base = self.base.strip('/')

    def _get_email(self, email):
        '''get an email (required) either provided by calling function or
           already set in client. Exit if not defined.

           Parameters
           ==========
           email: the email address associated with Twist
        '''
        email = email or self.email
        if not email:
            bot.exit("Email must provided to client, calling function, or in environment FREEGENES_TWIST_EMAIL.")
        return email

    def _set_email(self, email):
        '''look for FREEGENES_TWIST_EMAIL defined in environ
        '''
        self.email = os.environ.get('FREEGENES_TWIST_EMAIL', email)
        if not email:
            self.whoami()
        
    def _set_headers(self):
        '''set the headers to the default, meaning we provide an
           authorization token.
        '''
        self.headers = {
          "Accept": "application/json",
          "Accept-Encoding": "json",
          "Authorization": "JWT %s" % self.token,
          "X-End-User-Token": self.eutoken         
        }

    def _test_token(self):
        '''test that the token works - this function also ensures
           that the base is correct.
        '''
        if requests.head("%s" % self.base, headers=self.headers).status_code not in [200, 302]:
            bot.exit('Provided token is invalid.')


    # Specific API calls

    def get(self, url, headers=None, page=None, paginate=True):
        '''the default get, will use default headers if custom aren't defined.
           we take a partial url (e.g., /api/authors) and then add the base.

           Parameters
           ==========
           url: the url endpoint to query (without the http/s or domain)
           headers: if defined, don't use default headers.
           page: obtain a specific page of the result.
           paginate: obtain all pages after query (default is True)
        '''
        heads = headers or self.headers
        fullurl = "%s%s" %(self.base, url)

        # If we are provided a page
        if page:
            fullurl = "%s?page=%s" %(fullurl, page)
        response = requests.get(fullurl, headers=heads)

        # Return a successful response
        if response.status_code == 200:
 
            results = response.json()

            # Listings will have results, single entity not
            if "results" in results:
                results = results['results']

            # Are there pages?
            if paginate:
                while response.next:
                    results = results + self.get(url, headers, page=response.next)
            return results

        bot.exit("Error with %s, return value %s: %s" %(url, response.status_code, response.reason))


    # Endpoints

    def whoami(self):
        '''Returns the authentication APIClient username. 
           And email when EUT is supplied.
        '''
        result = self.get('/whoami/')
   
        # If the client did not provide an email, try getting it
        if not self.email and "email" in result:
            self.email = result["email"]
        return result

    def healthcheck(self):
        '''Returns the API healthcheck endpoint --> {'status': 'OK'}
        '''
        return self.get('/healthcheck/')

    def is_alive(self):
        '''Returns if the API is alive --> {'status': 'OK'}
        '''
        return self.get('/is_alive/')

    def version(self):
        '''Returns the API version information, including a dictionary with
           version, tag, branch, and git_commit.
        '''
        return self.get('/version/')

    # Accounts

    def accounts(self):
        '''Return accounts associated
        '''
        return self.get('/v1/accounts/')

    def account(self, account_id):
        '''Look up a specific account based on account id.
        '''
        return self.get('/v1/accounts/%s' % account_id)

    def account_contacts(self, account_id):
        '''Look up account contacts based on account id.
        '''
        return self.get('/v1/accounts/%s/contacts' % account_id)

    def account_users(self, account_id):
        '''Look up account users based on account id.
        '''
        return self.get('/v1/accounts/%s/users' % account_id)

    def account_prices(self, account_id):
        '''Look up account prices based on account id.
        '''
        return self.get('/v1/accounts/%s/prices' % account_id)

    # Catalog Items

    def catalog_items(self):
        '''Look up account prices based on account id.
        '''
        return self.get('/v1/catalog-items')

    def user(self, email=None):
        '''Look up user by email.
        '''
        email = self._get_email(email)
        return self.get('/v1/users/%s/' % email)

    def user_addresses(self, email=None):
        '''Look up user by email.
        '''
        email = self._get_email(email)
        return self.get('/v1/users/%s/addresses/' % email)

    def orders(self, email=None):
        '''Look up orders for a user based on email.
        '''
        email = self._get_email(email)
        return self.get('/v1/users/%s/orders/' % email)

    # Orders

    def order_items(self, sfdc_id, email=None):
        '''Look up order items for a user based on email 
           and the order sfdc_id.
        '''
        email = self._get_email(email)
        return self.get('/v1/users/%s/orders/%s/items' % (email, sfdc_id))

    def order_platemaps_by_barcode(self, sfdc_id, barcode, email=None):
        '''Look up order plate maps for a user based on email.
           sfdc_id and barcode.
        '''
        email = self._get_email(email)
        return self.get('/v1/users/%s/orders/%s/plate-maps/%s' % (email, sfdc_id, barcode))


    def order_platemaps_by_shipment(self, sfdc_id, shipment_id, email=None):
        '''Look up order plate maps for a user based on email.
           sfdc_id and shipment id.
        '''
        email = self._get_email(email)
        return self.get('/v1/users/%s/orders/%s/shipments/%s/plate-maps' % (email, sfdc_id, shipment_id))



from .fileio import ( 
    mkdir_p, 
    write_file, 
    write_json,
    read_file, 
    read_json
)

from .misc import ScopedEnvVar

from .terminal import (
    check_install, 
    format_container_name,
    get_installdir,
    get_singularity_version,
    get_singularity_version_info,
    remove_uri,
    run_command,
    stream_command,
    split_uri,
    which
)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from freegenes.main import Client
import shutil
import os


def _list_and_single(func):
    '''for a given endpoint that returns a list and a single view, 
       ensure that the list works to return 200 (and a list of results),
       and if we have a result, also query for a single entity. This is a
       simple test that just ensures that the endpoints are functioning.
    '''
    print('Testing %s' % func.__name__)
    results = func()
    assert isinstance(results, list)

    if len(results) > 0:
        result = func(uuid=results[0]['uuid'])
        assert isinstance(result, dict)
        assert "uuid" in result
        assert result['uuid'] == results[0]['uuid']


def test_endpoints():

    client = Client()
    _list_and_single(client.get_authors)
    _list_and_single(client.get_collections) 
    _list_and_single(client.get_containers)
    _list_and_single(client.get_distributions)
    _list_and_single(client.get_institutions)
    _list_and_single(client.get_modules)
    _list_and_single(client.get_operations)
    _list_and_single(client.get_orders)
    _list_and_single(client.get_organisms)
    _list_and_single(client.get_parts)
    _list_and_single(client.get_plans)
    _list_and_single(client.get_plates)
    _list_and_single(client.get_platesets)
    _list_and_single(client.get_protocols)
    _list_and_single(client.get_robots)




# Copyright (C) 2017-2019 Vanessa Sochat.

# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


import sys
import time
import threading
from random import choice

class Spinner:
    spinning = False
    delay = 0.1

    @staticmethod
    def spinning_cursor():
        while 1: 
            for cursor in '|/-\\':
                yield cursor

    @staticmethod
    def balloons_cursor():
        while 1: 
            for cursor in '. o O @ *':
                yield cursor

    @staticmethod
    def changing_arrows():
        while 1: 
            for cursor in '<^>v':
                yield cursor

    def select_generator(self, generator):
        if generator is None:
            generator = choice(['cursor',
                                'arrow',
                                'balloons'])

        return generator

    def __init__(self, delay=None, generator=None):
        generator = self.select_generator(generator)

        if generator == 'cursor':
            self.spinner_generator = self.spinning_cursor()
        elif generator == 'arrow':
            self.spinner_generator = self.changing_arrows()
        elif generator == 'balloons':
            self.spinner_generator = self.balloons_cursor()
            if delay is None:
                delay = 0.2
        else:
            self.spinner_generator = self.spinning_cursor()

        if delay and float(delay):
            self.delay = delay

    def run(self):
        while self.spinning:
            sys.stdout.write(next(self.spinner_generator))
            sys.stdout.flush()
            time.sleep(self.delay)
            sys.stdout.write('\b')
            sys.stdout.flush()

    def start(self):
        self.spinning = True
        threading.Thread(target=self.run).start()

    def stop(self):
        self.spinning = False
        time.sleep(self.delay)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from freegenes.version import __version__
from freegenes.logger import bot
import requests
import os

class Client(object):

    def __init__(self, token=None, base="https://freegenes.dev"):
 
        self._set_base(base)
        self._set_token(token)
        self._set_headers()
        self._test_token()

    def __repr__(self):
        return "[client][freegenes][%s]" % __version__

    def __str__(self):
        return "[client][freegenes][%s]" % __version__

    def _set_token(self, token):
        '''ensure that token provided, or FREEGENES_TOKEN is defined in environ
        '''
        self.token = os.environ.get('FREEGENES_TOKEN', token)
        if not self.token:
            bot.exit("You must provide a token or export FREEGENES_TOKEN.")

    def _set_base(self, base):
        '''look for FREEGENES_BASE defined in environ
        '''
        self.base = os.environ.get('FREEGENES_BASE', base)
        if self.base:
            self.base = self.base.strip('/')

    def _set_headers(self):
        '''set the headers to the default, meaning we provide an
           authorization token.
        '''
        self.headers = {
          "Accept": "application/json",
          "Authorization": "Token %s" % self.token
        }

    def _test_token(self):
        '''test that the token works - this function also ensures
           that the base is correct.
        '''
        if requests.head("%s" % self.base, headers=self.headers).status_code != 200:
            bot.exit('Provided token is invalid.')


    # Specific API calls

    def get(self, url, headers=None, page=None, paginate=True):
        '''the default get, will use default headers if custom aren't defined.
           we take a partial url (e.g., /api/authors) and then add the base.

           Parameters
           ==========
           url: the url endpoint to query (without the http/s or domain)
           headers: if defined, don't use default headers.
           page: obtain a specific page of the result.
           paginate: obtain all pages after query (default is True)
        '''
        heads = headers or self.headers
        fullurl = "%s%s" %(self.base, url)

        # If we are provided a page
        if page:
            fullurl = "%s?page=%s" %(fullurl, page)
        response = requests.get(fullurl, headers=heads)

        # Return a successful response
        if response.status_code == 200:
 
            results = response.json()

            # Listings will have results, single entity not
            if "results" in results:
                results = results['results']

            # Are there pages?
            if paginate:
                while response.next:
                    results = results + self.get(url, headers, page=response.next)
            return results

        bot.exit("Error with %s, return value %s: %s" %(url, response.status_code, response.reason))


    # Endpoints

    def get_entity(self, name, uuid=None):
        '''return a single entity if a uuid is provided, otherwise a list

           Parameters
           ==========
           uuid: the unique resource identifier of the entity
        '''
        if uuid:
            return self.get('/api/%s/%s' % (name, uuid))
        return self.get('/api/%s' % name)


    def get_authors(self, uuid=None):
        return self.get_entity('authors', uuid)

    def get_collections(self, uuid=None):
        return self.get_entity('collections', uuid)

    def get_containers(self, uuid=None):
        return self.get_entity('containers', uuid)

    def get_distributions(self, uuid=None):
        return self.get_entity('distributions', uuid)

    def get_institutions(self, uuid=None):
        return self.get_entity('institutions', uuid)

    def get_modules(self, uuid=None):
        return self.get_entity('modules', uuid)

    def get_operations(self, uuid=None):
        return self.get_entity('operations', uuid)

    def get_orders(self, uuid=None):
        return self.get_entity('orders', uuid)

    def get_organisms(self, uuid=None):
        return self.get_entity('organisms', uuid)

    def get_parts(self, uuid=None):
        return self.get_entity('parts', uuid)

    def get_plans(self, uuid=None):
        return self.get_entity('plans', uuid)

    def get_plates(self, uuid=None):
        return self.get_entity('plates', uuid)

    def get_platesets(self, uuid=None):
        return self.get_entity('platesets', uuid)

    def get_protocols(self, uuid=None):
        return self.get_entity('protocols', uuid)

    def get_robots(self, uuid=None):
        return self.get_entity('robots', uuid)



# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os

def setEnvVar(name, value):
    '''Set or unset an environment variable

    name -- Name of the variable to set
    value -- Value to use or None to clear
    '''
    if value is None:
        if name in os.environ:
            del os.environ[name]
    else:
        os.environ[name] = value

class ScopedEnvVar(object):
    '''Temporarly change an environment variable

    Usage:
        with ScopedEnvVar("FOO", "bar"):
            print(os.environ["FOO"]) # "bar"
        print(os.environ["FOO"]) # <oldvalue>
    '''
    
    def __init__(self, name, value):
        '''Create the scoped environment variable object

        name -- Name of the variable to set
        value -- Value to use or None to clear
        '''
        self.name = name
        self.value = value
        self.oldValue = None
    
    def __enter__(self):
        self.oldValue = os.environ.get(self.name)
        setEnvVar(self.name, self.value)
        return self
    
    def __exit__(self, ex_type, ex_value, traceback):
        setEnvVar(self.name, self.oldValue)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from freegenes.main.twist import Client
from .shell import (
    ipython,
    python,
    run_bpython
)

def main(args, options, parser):

    # Choose executor based on what is available 
    lookup = {'ipython': ipython,
              'python': python,
              'bpython': run_bpython}

    shells = ['ipython', 'python', 'bpython']

    # Prepare client
    client = Client()

    # Otherwise present order of liklihood to have on system
    for shell in shells:
        try:
            return lookup[shell](client)
        except ImportError:
            pass



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from setuptools import setup, find_packages
import os

################################################################################
# HELPER FUNCTIONS #############################################################
################################################################################

def get_lookup():
    '''get version by way of the version file, returns a 
       lookup dictionary with several global variables without
       needing to import spython
    '''
    lookup = dict()
    version_file = os.path.join('freegenes', 'version.py')
    with open(version_file) as filey:
        exec(filey.read(), lookup)
    return lookup

def get_requirements(lookup=None):
    '''get_requirements reads in requirements and versions from
       the lookup obtained with get_lookup'''

    if lookup is None:
        lookup = get_lookup()

    install_requires = []
    for module in lookup['INSTALL_REQUIRES']:
        module_name = module[0]
        module_meta = module[1]
        if "exact_version" in module_meta:
            dependency = "%s==%s" % (module_name, module_meta['exact_version'])
        elif "min_version" in module_meta:
            if module_meta['min_version'] is None:
                dependency = module_name
            else:
                dependency = "%s>=%s" % (module_name, module_meta['min_version'])
        install_requires.append(dependency)
    return install_requires



# Make sure everything is relative to setup.py
install_path = os.path.dirname(os.path.abspath(__file__)) 
os.chdir(install_path)

# Get version information from the lookup
lookup = get_lookup()
VERSION = lookup['__version__']
NAME = lookup['NAME']
AUTHOR = lookup['AUTHOR']
AUTHOR_EMAIL = lookup['AUTHOR_EMAIL']
PACKAGE_URL = lookup['PACKAGE_URL']
KEYWORDS = lookup['KEYWORDS']
DESCRIPTION = lookup['DESCRIPTION']
LICENSE = lookup['LICENSE']
with open('README.md') as readme:
    LONG_DESCRIPTION = readme.read()

################################################################################
# MAIN #########################################################################
################################################################################


if __name__ == "__main__":

    INSTALL_REQUIRES = get_requirements(lookup)
    TESTS_REQUIRES = get_requirements(lookup)

    setup(name=NAME,
          version=VERSION,
          author=AUTHOR,
          author_email=AUTHOR_EMAIL,
          maintainer=AUTHOR,
          maintainer_email=AUTHOR_EMAIL,
          packages=find_packages(), 
          include_package_data=True,
          zip_safe=False,
          url=PACKAGE_URL,
          license=LICENSE,
          description=DESCRIPTION,
          long_description=LONG_DESCRIPTION,
          long_description_content_type="text/markdown",
          keywords=KEYWORDS,
          setup_requires=["pytest-runner"],
          tests_require=TESTS_REQUIRES,
          install_requires=INSTALL_REQUIRES,
          classifiers=[
              'Intended Audience :: Science/Research',
              'Intended Audience :: Developers',
              'Programming Language :: Python',
              'Topic :: Software Development',
              'Topic :: Scientific/Engineering',
              'Operating System :: Unix',
              'Programming Language :: Python :: 3',
          ],
          entry_points={'console_scripts': ['freegenes=freegenes.client:main']})



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from freegenes.main import Client

def main(args, options, parser):

    # Choose executor based on what is available 
    lookup = {'ipython': ipython,
              'python': python,
              'bpython': run_bpython}

    shells = ['ipython', 'python', 'bpython']

    # Prepare client
    client = Client()

    # Otherwise present order of liklihood to have on system
    for shell in shells:
        try:
            return lookup[shell](client)
        except ImportError:
            pass

def ipython(client):
    '''give the user an ipython shell
    '''
    try:
        from IPython import embed
    except ImportError:
        return python(client)

    embed(using=False)

def run_bpython(client):
    '''give the user a bpython shell
    '''
    try:
        import bpython
    except ImportError:
        return python(client)

    bpython.embed(locals_={'client': client})

def python(client):
    '''give the user a python shell
    '''
    import code
    code.interact(local={"client":client})
